Public Class frmRoman2

    Private Sub optDecimalToRoman_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles optDecimalToRoman.CheckedChanged

        If optDecimalToRoman.Checked Then
            lblPrompt.Text = "Enter a number between 1 and 3999:"
            lblOutputDesc.Text = "Roman Numeral Equivalent:"
            txtInput.Text = ""
            txtInput.MaxLength = 4
            lblOutput.Text = ""
        End If

    End Sub

    Private Sub optRomanToDecimal_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles optRomanToDecimal.CheckedChanged

        If optRomanToDecimal.Checked Then
            lblPrompt.Text = "Enter a Roman Numerals number between I and MMMCMXCIX:"
            lblOutputDesc.Text = "Decimal Equivalent:"
            txtInput.Text = ""
            txtInput.MaxLength = 16
            lblOutput.Text = ""
        End If

    End Sub

    Private Sub btnConvert_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnConvert.Click

        Dim strRomanNumeralString As String
        Dim strInvalidRomanInputMsg As String

        If optDecimalToRoman.Checked = True Then
            If Val(txtInput.Text) < 1 Or Val(txtInput.Text) > 3999 Then
                MsgBox("Number out of range. Please enter a number between 1 and 3999.", _
                       MsgBoxStyle.Exclamation, _
                       "Number Out of Range")
                txtInput.Focus()
                Exit Sub
            End If
            lblOutput.Text = ConvertToRoman((txtInput.Text))
        Else
            strRomanNumeralString = txtInput.Text
            If Not ValidRomanInput(strRomanNumeralString, strInvalidRomanInputMsg) Then
                MsgBox(strInvalidRomanInputMsg, _
                       MsgBoxStyle.Exclamation, _
                       "Invalid Roman Numeral String")
                txtInput.Focus()
                Exit Sub
            End If
            lblOutput.Text = ConvertToDecimal(txtInput.Text)
        End If

    End Sub

    Private Sub btnExit_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnExit.Click
        Me.Close()
    End Sub

    Private Sub txtInput_GotFocus(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles txtInput.GotFocus
        txtInput.SelectAll()
    End Sub

    Private Sub txtInput_KeyPress(ByVal sender As System.Object, ByVal e As System.Windows.Forms.KeyPressEventArgs) Handles txtInput.KeyPress

        If Asc(e.KeyChar) < 32 Then
            e.Handled = False
            Exit Sub
        End If

        If optDecimalToRoman.Checked = True Then
            e.Handled = Not (Char.IsDigit(e.KeyChar))
        Else
            ' Note: The CharacterCasing property of txtInput was set to Upper at
            '       design time.
            e.Handled = (InStr("MDCLXVI", Char.ToUpper(e.KeyChar)) = 0)
        End If

    End Sub

    Private Sub txtInput_TextChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles txtInput.TextChanged
        lblOutput.Text = ""
    End Sub

    Private Function ConvertToRoman(ByRef pstrDecimalNumber As String) As String

        Const strPOS_VAL As String = "IXCM"
        Const strFIVE_VAL As String = "VLD"

        Dim strRoman As String
        Dim strCurrRomanPos As String
        Dim strLetter1 As String
        Dim strLetter2 As String
        Dim intCurrPos As Short
        Dim intDigit As Short
        Dim intDigitPos As Short

        intCurrPos = 1
        strRoman = ""

        For intDigitPos = Len(pstrDecimalNumber) To 1 Step -1
            intDigit = Val(Mid(pstrDecimalNumber, intDigitPos, 1))
            strCurrRomanPos = Mid(strPOS_VAL, intCurrPos, 1)
            Select Case intDigit
                Case 9
                    strLetter1 = strCurrRomanPos
                    strLetter2 = Mid(strPOS_VAL, intCurrPos + 1, 1)
                Case Is > 4
                    strLetter1 = Mid(strFIVE_VAL, intCurrPos, 1)
                    strLetter2 = New String(strCurrRomanPos, intDigit - 5)
                Case 4
                    strLetter1 = strCurrRomanPos
                    strLetter2 = Mid(strFIVE_VAL, intCurrPos, 1)
                Case Else
                    strLetter1 = New String(strCurrRomanPos, intDigit)
                    strLetter2 = ""
            End Select
            strRoman = strLetter1 & strLetter2 & strRoman
            intCurrPos = intCurrPos + 1
        Next

        ConvertToRoman = strRoman

    End Function

    Private Function ConvertToDecimal(ByRef pstrRomanNumeral As String) As String

        Dim aintRomanValues() As Short
        Dim intInputLen As Short
        Dim intX As Short
        Dim intSum As Short

        intInputLen = Len(pstrRomanNumeral)

        If intInputLen = 0 Then
            ConvertToDecimal = CStr(0)
            Exit Function
        End If

        ReDim aintRomanValues(intInputLen)

        For intX = 1 To intInputLen
            Select Case Mid(pstrRomanNumeral, intX, 1)
                Case "M" : aintRomanValues(intX) = 1000
                Case "D" : aintRomanValues(intX) = 500
                Case "C" : aintRomanValues(intX) = 100
                Case "L" : aintRomanValues(intX) = 50
                Case "X" : aintRomanValues(intX) = 10
                Case "V" : aintRomanValues(intX) = 5
                Case "I" : aintRomanValues(intX) = 1
            End Select
        Next

        For intX = 1 To intInputLen
            If intX = intInputLen Then
                intSum = intSum + aintRomanValues(intX)
            Else
                If aintRomanValues(intX) >= aintRomanValues(intX + 1) Then
                    intSum = intSum + aintRomanValues(intX)
                Else
                    intSum = intSum - aintRomanValues(intX)
                End If
            End If
        Next

        ConvertToDecimal = CStr(intSum)

    End Function

    Private Function ValidRomanInput(ByVal pstrRN As String, _
                                     ByRef pstrMsg As String) _
    As Boolean

        ValidRomanInput = False ' Guilty until proven innocent!

        ' a 'D', 'L', or 'V' may only appear at most once in the string
        If GetSubstringCount(pstrRN, "D") > 1 _
        Or GetSubstringCount(pstrRN, "L") > 1 _
        Or GetSubstringCount(pstrRN, "V") > 1 _
        Then
            pstrMsg = "'D', 'L', or 'V' may only appear at most once."
            Exit Function
        End If

        ' no more than 3 consecutive Ms, Cs, Xs or Is:
        If InStr(pstrRN, "MMMM") > 0 _
        Or InStr(pstrRN, "CCCC") > 0 _
        Or InStr(pstrRN, "XXXX") > 0 _
        Or InStr(pstrRN, "IIII") > 0 _
        Then
            pstrMsg = "'M', 'C', 'X', or 'I' may appear no more than three times in a row."
            Exit Function
        End If

        ' Outright illegal sequences:
        ' --------------------------
        ' Only I, X, and C can be used for subtraction (V, L, and D cannot). Therefore, the
        ' following pairs of letter are invalid: VX, VL, VC, VD, VM, LC, LD, LM, DM.
        ' When subtracting, the value of the letter being subtracted from cannot be more than
        ' 10 times the value of letter being used for subtraction. Therefore, the following pairs
        ' of letters are invalid: IL, IC, ID, IM, XD, XM.
        If InStr(pstrRN, "IL") > 0 _
        Or InStr(pstrRN, "IC") > 0 _
        Or InStr(pstrRN, "ID") > 0 _
        Or InStr(pstrRN, "IM") > 0 _
        Or InStr(pstrRN, "XD") > 0 _
        Or InStr(pstrRN, "XM") > 0 _
        Or InStr(pstrRN, "VX") > 0 _
        Or InStr(pstrRN, "VL") > 0 _
        Or InStr(pstrRN, "VC") > 0 _
        Or InStr(pstrRN, "VD") > 0 _
        Or InStr(pstrRN, "VM") > 0 _
        Or InStr(pstrRN, "LC") > 0 _
        Or InStr(pstrRN, "LD") > 0 _
        Or InStr(pstrRN, "LM") > 0 _
        Or InStr(pstrRN, "DM") > 0 _
        Then
            pstrMsg = "The Roman Numeral string contains an illegal sequence of characters."
            Exit Function
        End If

        ' Other illegal sequences:

        ' Once a letter has been used as a subtraction modifier, that letter cannot appear again
        ' in the string, unless that letter itself is subtracted from. For example, CDC is not
        ' valid (you would be subtracting 100 from 500, then adding it right back) -
        ' but CDXC (for 490) is valid. Similarly, XCX is not valid, but XCIX is.
        ' To summarize:
        ' C cannot follow CM or CD except in case of XC.
        ' X cannot follow XC or XL except in the case of IX.
        If AFollowsBInCExceptD("C", "CD", pstrRN, "XC") Then pstrMsg = "'C' cannot follow 'CD' except for 'XC'." : Exit Function
        If AFollowsBInCExceptD("C", "CM", pstrRN, "XC") Then pstrMsg = "'C' cannot follow 'CD' except for 'XC'." : Exit Function
        If AFollowsBInCExceptD("X", "XL", pstrRN, "IX") Then pstrMsg = "'X' cannot follow 'XL' except for 'IX'." : Exit Function
        If AFollowsBInCExceptD("X", "XC", pstrRN, "IX") Then pstrMsg = "'X' cannot follow 'XL' except for 'IX'." : Exit Function
        If AFollowsBInC("I", "IV", pstrRN) Then pstrMsg = "'I' cannot follow 'IV'." : Exit Function
        If AFollowsBInC("I", "IX", pstrRN) Then pstrMsg = "'I' cannot follow 'IV'." : Exit Function

        ' Once a letter has been subtracted from, neither it nor the next lowest multiple of 5 may
        ' appear again in the string - so neither X nor V can follow IX, neither C nor L may follow
        ' XC, and neither M nor D may follow CM.
        If AFollowsBInC("X", "IX", pstrRN) Then pstrMsg = "'X' cannot follow 'IX'." : Exit Function
        If AFollowsBInC("V", "IX", pstrRN) Then pstrMsg = "'V' cannot follow 'IX'." : Exit Function
        If AFollowsBInC("C", "XC", pstrRN) Then pstrMsg = "'C' cannot follow 'XC'." : Exit Function
        If AFollowsBInC("L", "XC", pstrRN) Then pstrMsg = "'L' cannot follow 'XC'." : Exit Function
        If AFollowsBInC("M", "CM", pstrRN) Then pstrMsg = "'M' cannot follow 'CM'." : Exit Function
        If AFollowsBInC("D", "CM", pstrRN) Then pstrMsg = "'D' cannot follow 'CM'." : Exit Function

        ' A letter cannot be used as a subtraction modifier if that letter, or the next highest
        ' multiple of 5, appears previously in the string - so IV or IX cannot follow I or V,
        ' XL or XC cannot follow X or L, and CD or CM cannot follow C or D.
        If AFollowsBInC("IV", "I", pstrRN) Then pstrMsg = "'IV' cannot follow 'I'." : Exit Function
        If AFollowsBInC("IX", "I", pstrRN) Then pstrMsg = "'IX' cannot follow 'I'." : Exit Function
        If AFollowsBInC("IX", "V", pstrRN) Then pstrMsg = "'IX' cannot follow 'V'." : Exit Function
        If AFollowsBInC("XL", "X", pstrRN) Then pstrMsg = "'XL' cannot follow 'X'." : Exit Function
        If AFollowsBInC("XC", "X", pstrRN) Then pstrMsg = "'XC' cannot follow 'X'." : Exit Function
        If AFollowsBInC("XC", "L", pstrRN) Then pstrMsg = "'XC' cannot follow 'L'." : Exit Function
        If AFollowsBInC("CD", "C", pstrRN) Then pstrMsg = "'CD' cannot follow 'C'." : Exit Function
        If AFollowsBInC("CM", "C", pstrRN) Then pstrMsg = "'CM' cannot follow 'C'." : Exit Function
        If AFollowsBInC("CM", "D", pstrRN) Then pstrMsg = "'CM' cannot follow 'D'." : Exit Function

        ' If we make it here, the Roman Numeral input was valid ...
        ValidRomanInput = True

    End Function

    Private Function GetSubstringCount(ByVal pstrMainString As String, _
                                       ByVal pstrSubstring As String) _
    As Integer

        Dim lngX As Integer
        Dim lngY As Integer

        If pstrMainString = "" Then
            GetSubstringCount = 0
        Else
            lngX = InStr(1, pstrMainString, pstrSubstring, CompareMethod.Binary)
            If lngX = 0 Then
                GetSubstringCount = 0
            Else
                lngX = 0
                For lngY = 1 To Len(pstrMainString)
                    If Mid(pstrMainString, lngY, Len(pstrSubstring)) = pstrSubstring Then
                        lngX = lngX + 1
                    End If
                Next lngY
                GetSubstringCount = lngX
            End If
        End If

    End Function

    Private Function AFollowsBInC(ByRef pstrA As String, _
                                  ByRef pstrB As String, _
                                  ByRef pstrC As String) _
    As Boolean

        Dim lngTestPos As Integer

        lngTestPos = InStr(pstrC, pstrB)

        If lngTestPos > 0 Then
            If InStr(lngTestPos + Len(pstrB), pstrC, pstrA, CompareMethod.Text) Then
                AFollowsBInC = True
            Else
                AFollowsBInC = False
            End If
        Else
            AFollowsBInC = False
        End If

    End Function

    Private Function AFollowsBInCExceptD(ByVal pstrA As String, _
                                         ByVal pstrB As String, _
                                         ByVal pstrC As String, _
                                         ByVal pstrD As String) _
    As Boolean

        Dim lngTestPos As Long
        Dim strRemChars As String

        lngTestPos = InStr(pstrC, pstrB)

        If lngTestPos = 0 Then
            AFollowsBInCExceptD = False
            Exit Function
        End If

        strRemChars = Mid(pstrC, lngTestPos + Len(pstrB))
        strRemChars = Replace(strRemChars, pstrD, "")
        If InStr(strRemChars, pstrA) > 0 Then
            AFollowsBInCExceptD = True
        Else
            AFollowsBInCExceptD = False
        End If

    End Function


End Class
